(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[5405],{63878:function(n,t,u){"use strict";u.r(t);var a=u(82684),e=u(34376),i=u(41374);t.default=function(){var n,t=(0,e.useRouter)(),u=t.asPath,s=u.split("?")[0],_=i.ZP.status.list().data,c=null===_||void 0===_||null===(n=_.status)||void 0===n?void 0:n.is_instance_manager,o=u;"/"===s&&(o=c?"/manage":"/pipelines"),(0,a.useEffect)((function(){_&&t.replace(o)}),[_])}},48312:function(n,t,u){(window.__NEXT_P=window.__NEXT_P||[]).push(["/",function(){return u(63878)}])}},function(n){n.O(0,[4259,1374,9774,2888,179],(function(){return t=48312,n(n.s=t);var t}));var t=n.O();_N_E=t}]);