self.__BUILD_MANIFEST=function(s,e,i,c,n,p,t,a,l,u,d,b,r,k,f,h,j,g,o,_,m,E,O,A,B,L,D,T,C){return{__rewrites:{beforeFiles:[],afterFiles:[{source:"/CLOUD_NOTEBOOK_BASE_PATH_PLACEHOLDER_/datasets",destination:j},{source:"/CLOUD_NOTEBOOK_BASE_PATH_PLACEHOLDER_",destination:j}],fallback:[]},"/":[s,i,"static/chunks/pages/index-242a33f6b14ebd84.js"],"/_error":["static/chunks/pages/_error-235304e5badb19eb.js"],"/datasets":[s,e,c,n,A,"static/chunks/pages/datasets-1c67cd14d15264e6.js"],"/datasets/[...slug]":[g,"static/chunks/8a28b14e-b845f39223877713.js",s,e,c,l,d,r,k,o,B,"static/chunks/17-6f5e28da6ea3bc27.js",n,t,a,f,_,L,A,"static/chunks/pages/datasets/[...slug]-13a0106fe9628467.js"],"/manage":[s,e,c,i,n,p,a,"static/chunks/pages/manage-4239274f6872a6d2.js"],"/pipeline-runs":[s,e,c,i,n,p,a,m,"static/chunks/pages/pipeline-runs-c031de6c051f00d1.js"],"/pipelines":[s,e,c,i,n,p,"static/chunks/pages/pipelines-f97c8f9de1892b22.js"],"/pipelines/[pipeline]":["static/chunks/pages/pipelines/[pipeline]-0c6dc5356df3d124.js"],"/pipelines/[pipeline]/edit":[g,s,e,c,l,b,d,r,k,E,o,O,D,B,"static/chunks/1869-76da4d07a68399ef.js",i,n,p,t,a,f,h,_,L,"static/chunks/pages/pipelines/[pipeline]/edit-087bc8c6ebf2e2d8.js"],"/pipelines/[pipeline]/logs":[s,e,c,b,O,T,i,n,p,t,a,u,"static/chunks/pages/pipelines/[pipeline]/logs-634f891a54082323.js"],"/pipelines/[pipeline]/monitors":[s,e,c,l,k,i,n,p,t,u,C,"static/chunks/pages/pipelines/[pipeline]/monitors-00ea20e16b43e3c0.js"],"/pipelines/[pipeline]/monitors/block-runs":[s,e,c,l,k,i,n,p,t,a,u,C,"static/chunks/pages/pipelines/[pipeline]/monitors/block-runs-b3704cf56cf921c1.js"],"/pipelines/[pipeline]/monitors/block-runtime":[g,s,e,c,l,k,"static/chunks/9832-c8b8970bb522f302.js",i,n,p,t,a,u,_,"static/chunks/pages/pipelines/[pipeline]/monitors/block-runtime-009a37797d622ffb.js"],"/pipelines/[pipeline]/runs":[s,e,c,b,d,E,i,n,p,t,a,u,h,m,"static/chunks/pages/pipelines/[pipeline]/runs-bc5b1982fc45d40a.js"],"/pipelines/[pipeline]/runs/[run]":[s,e,c,l,b,d,r,o,i,n,p,t,u,f,h,"static/chunks/pages/pipelines/[pipeline]/runs/[run]-c8108a401e5aa314.js"],"/pipelines/[pipeline]/triggers":[s,e,c,l,b,d,r,i,n,p,t,u,f,h,"static/chunks/pages/pipelines/[pipeline]/triggers-86e41a40f8e2ae5d.js"],"/pipelines/[pipeline]/triggers/[...slug]":[s,e,c,l,b,d,r,E,T,i,n,p,t,a,u,f,h,m,"static/chunks/pages/pipelines/[pipeline]/triggers/[...slug]-3b3d511f6e1281fd.js"],"/terminal":[s,e,O,D,i,"static/chunks/pages/terminal-c248ec9e65efd81b.js"],sortedPages:["/","/_app","/_error",j,"/datasets/[...slug]","/manage","/pipeline-runs","/pipelines","/pipelines/[pipeline]","/pipelines/[pipeline]/edit","/pipelines/[pipeline]/logs","/pipelines/[pipeline]/monitors","/pipelines/[pipeline]/monitors/block-runs","/pipelines/[pipeline]/monitors/block-runtime","/pipelines/[pipeline]/runs","/pipelines/[pipeline]/runs/[run]","/pipelines/[pipeline]/triggers","/pipelines/[pipeline]/triggers/[...slug]","/terminal"]}}("static/chunks/4259-f723eee7c5c19f8e.js","static/chunks/2212-f7752030be6ab268.js","static/chunks/1374-aba9a59280f0ef4f.js","static/chunks/7689-30032583a188b1a4.js","static/chunks/5763-e8198fe40a779f63.js","static/chunks/6792-30e5e6ee58f37731.js","static/chunks/1273-0305ef0e27954a99.js","static/chunks/9898-98e596a737fbb34b.js","static/chunks/6674-63776cbaecd7c1d9.js","static/chunks/8965-ffd46d62946ef37e.js","static/chunks/1774-aa51ef1da7217ff9.js","static/chunks/4804-01a10103ebe26ca8.js","static/chunks/3668-3b07d0117a5db904.js","static/chunks/2714-1e79e9f2e998b544.js","static/chunks/2151-11e7139c32bb7ab4.js","static/chunks/5703-ab4054a03e154abf.js","/datasets","static/chunks/29107295-989a0767a635d9d5.js","static/chunks/2125-d859c4773f5f76a8.js","static/chunks/3461-64870781b97c10bd.js","static/chunks/4846-51b5abd3ebbdf822.js","static/chunks/7619-11107d4d093f5d51.js","static/chunks/9565-28e1c30511c95c2d.js","static/chunks/3887-13f5cc240b3acfc2.js","static/chunks/5650-7bd2f076f244e576.js","static/chunks/9223-296d89ed2dd314a4.js","static/chunks/6085-692d2f784c0504f2.js","static/chunks/4495-aaa900f908a8a255.js","static/chunks/7400-a578d543bd286bdb.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();